/* ********************************************************************
 * ******************************************************************** 
 * ****								   **** 
 * ****								   **** 
 * ****	      	EPICURE Beamline Control System			   **** 
 * ****								   **** 
 * ****        	Programmer's Service Program		           **** 
 * ****								   **** 
 * ****								   **** 
 * ******************************************************************** 
 * ******************************************************************** 
 *
 *
 *
 * FUNCTIONAL DESCRIPTION:
 */
/*+ EXTRACTOR
 *
 *\subnam	EXTRACTOR
 *\subcal	$EXTRACTOR/qualifiers  filename
 *\subtxt	This program will scan a code file looking for properly delimited
 *	blocks of documentation.  An example of "proper delimiters" is
 *	\literal
 *	C+ ROUTINE_NAME
 *	C  embedded formatting flags, and text
 *	C-
 *	\endliteral
 *	
 *	To run the extractor, define the following symbol (or one like it):
 *	\literal
 *	   extract :== $RDCS$EXE:EXTRACTOR
 *	\endliteral
 *	and then type
 *	\literal
 *	   extract /qualifiers    file-spec
 *	\endliteral
 *	as per directions in write-up.
 *  
 *	The EXTRACTOR will produce up  to  four types of output.  One
 *	contains LaTeX formatting commands  and  can be processed and then
 *	printed on a laser  printer.    Another contains DSR commands with
 *	level numbers in position so  that  it  can  be included in a help
 *	library after processing.  A third will also contain DSR commands,
 *	but will not list the level  numbers.  It is designed for printout
 *	on a generic or "vanilla"  printer.    The fourth output option is
 *	designed for backward  compatibility  with existing programs which
 *	contain the +/- block extracting flags, but no embedded formatting
 *	flags.  In this  case,  the  extractor  will extract the delimited
 *	block, optionally remove the  comment  delimiters at the beginning
 *	of each line, and just print the text as is to a file, without any
 *	formatting.
 *
 *	Please note that, unless  the  option  is selected to produce
 *	only the PLAIN file (the last desribed above), it is considered an
 *	error if the file does not have embedded formatting flags.
 *
 *	Command Line Parameter and Qualifiers:
 *\arglist
 *\argn	filename         Specifies what file(s) the  EXTRACTOR  is to work on.
 *                       If a file extension is  omitted, or is .LIS, the file
 *                       is assumed to contain  a  list  of filenames on which
 *                       the  EXTRACTOR  is   to   operate.    Otherwise,  the
 *                       extension  must  be   one   of  the  recognized  set,
 *                       currently including:
 *			\intlist
 *			\intn	.C
 *			\intn	.COM
 *			\intn	.FOR
 *			\intn	.FTN
 *			\intn	.H
 *			\intn	.MAC
 *			\intn	.MAR
 *			\intn	.PAS
 *			\intn	.Z8K
 *			\intn	.Z80
 *			\intn	.68K
 *			\intlend
 *                       If  filename is  omitted  entirely  the  program will
 *                       create  a  temporary  file  containing  the directory
 *                       listing of the current default directory, and process
 *                       any files with recognized extensions.
 *\argn	/ALPHABETIZE   (or /NOALPHABETIZE)  Specifies whether separate 
 *			extractable blocks which the extractor finds are to
 *			be ordered alphabetically or left in the order they 
 *			were found.  /ALPHABETIZE is the default.
 *\argn	/EMBEDDED      (or /NOEMBEDDED)  If /EMBEDDED is selected, 
 *                       the file is assumed to have
 *                       embedded formatting commands.   It will be considered
 *                       an error if none are  found.  /NOEMBEDDED may only be
 *                       selected  with  PLAIN, HELP, EPIC, or CLIB  processing
 *                       options; it may not  be  selected with LATEX, RUNOFF,
 *                       or any combination  which  includes these.  /EMBEDDED
 *                       is the default; however,  selecting /NOREMOVE forces 
 *                       /NOEMBEDDED.
 *\argn	/LOG          (or /NOLOG) This option  selects  whether  information 
 *                       about what
 *                       headers were processed  and  what  errors were found,
 *                       etc. should be saved  to file EXTRACTED.LOG.  Default
 *                       is to LOG everything.
 *\argn	/QUIET         (or /NOQUIET - DEFAULT)  This option  selects  
 *                       whether  information about what
 *                       headers were  processed  and  what  errors were found
 *                       should be printed to the terminal.  The default is to
 *                       print them.
 *\argn	/REMOVE        (or /NOREMOVE)  If /REMOVE is  selected,  the  
 *                       comment markers at the
 *                       start of each line  of the header (where appropriate)
 *                       are removed from the output text.  /NOREMOVE may only
 *                       be selected  with  PLAIN or EPIC processing options;
 *                       it may not  be  selected with LATEX, RUNOFF, HELP,
 *                       or any combination which  includes these.  /REMOVE is
 *                       the default if nothing is selected.
 *\argn	/LEVEL=n     Specifies  level  of   documentation   blocks  to  be
 *                       extracted.  This refers to  the number or space which
 *                       comes after the + sign  in the block delimiters which
 *                       flag  the  beginning  of  a  new  block.   Acceptable
 *                       choices for n:
 *			\intlist
 *			\intn	n        where n is between  1 and 9.  Space
 *                                         is interpretted the same as a 1.
 *			\intn	n-m      n,m   both   between   1   and   9,
 *                                         inclusive.  Example   3-5.
 *			\intn	nTOm     n,m   both   between   1   and   9,
 *                                         inclusive.  Example   3TO5.
 *			\intn	(n,m..)  List   of   integers   enclosed  in
 *                                         parens.
 *			\intn	ALL      All blocks, which means 1 through 9
 *                                         and includes "space".
 *			\intlend
 *\argn	/OUTPUT=filename Specifies the file to be  written if only one type of
 *                       output file is selected.    If  more than one type of
 *                       output is to be produced (see PROCESSING) this file
 *			 extension is ignored, and default file extensions are
 *			 used.  They are:
 *			\intlist
 *			\intn	filename.HLP     help file output, if NOEMBEDDED
 *			\intn	filename.HLPRNO  help file output which must be
 *						 RUNOFFed
 *			\intn	filename.RNO     runoff-able output
 *			\intn	filename.TEX     LaTeX-able output
 *			\intn	filename.TXT     plain output
 *			\intlend
 *\argn	/PROCESSING  Specifies  what  type  of  output  files  are  to  be
 *                       produced, and what type of processing must be done on
 *                       embedded commands.  Valid options are /PROCESSING=:
 *			\intlist
 *			\intn	PLAIN    Documentation block  is  written as
 *                                         is to a  file,  once extracted from
 *                                         code.
 *			\intn	RUNOFF   Substitutes  runoff   commands  for
 *                                         emmedded    formatting    commands.
 *                                         Output (file.RNO)  must be RUNOFFed
 *                                         before printing.
 *			\intn	HELP     Substitutes runoff commands for any
 *                                         embedded commands present.  Inserts
 *                                         help  level  numbers  where needed.
 *                                         Output (file.HLP)  may  need  to be
 *                                         RUNOFFed if  embedded commands were
 *                                         present.  (Embedded commands should
 *                                         be  present  in  all  newly written
 *                                         code.)
 *			\intn	LATEX    Substitutes   LaTeX   commands  for
 *                                         embedded    formatting    commands.
 *                                         Output  must   be   LATEXed  before
 *                                         printing (to a laser printer).
 *					   (DEFAULT)
 *			\intn	YES      Same as (RUNOFF,HELP,LATEX).
 *			\intn	ALL      Same as (RUNOFF,HELP,LATEX,PLAIN).
 *			\intn	CLIB     Same as (PLAIN,HELP).
 *			\intn	EPIC     Same as PLAIN.
 *			\intlend
 *                       A  list  of  options  may  also  be  specified within
 *                       parens, as /PROCESSING=(LATEX,RUNOFF).
 *\argn	/HELP_LEVEL=n	Specifies the level number to be output in the help
 *			text file.  Valid only if processing option HELP is
 *			selected.
 *
 *	
 *\argn	INTERNALLY:
 *          Embedded formatting flags recognized by this routine are as follows:
 *	\intlist
 *	\intn	\subnam      Signals that the next  single  word is the 
 *				subroutine or function name.
 *	\intn	\subcal      The text which follows is the normal calling 
 *				sequence for the function.
 *	\intn	\subtxt      Signals the end of the '\subcal' phrase and 
 *				the start of  the  descriptive  text  about  
 *				the  subroutine or function.
 *	\intn	\arglist     Begins the list  of  argument  descriptions.  
 *				Must be included before any '\argn' flag.
 *	\intn	\argn         Signals that the  next  single  word  is the 
 *				argument name.  Following the argument name 
 *				is the descriptive text about that argument.
 *	\intn	\intlist     Signals a  list  'internal'  to  the  argument 
 *				names.  Suggested usage  is  for  listing  
 *				possible error returns.  Internal lists 
 *				may not be nested as currently defined.
 *	\intn	\intn         The next single word is  the keyword for 
 *				the internal list item. Text for the list
 *				item  follows the keyword.
 *	\intn	\intlend     Required syntax to  signify  the  end  of 
 *				an internal list of items.
 *	\intn	\arglend     Required syntax to signify the end of an
 *				internal list of items.
 *	\intn	\literal     Prints following line exactly as typed.  
 *	\intn   \literal     Prints following lines  exactly  as  typed.  
 *			     Intended for diagramming.   No  text  may  be 
 *			     included on this line.   Diagrams  should  start  
 *			     in  the first column after any comment  delimiters.
 *			     (Asterisks in Pascal or C are assumed to be comments
 *			     in column 1 or 2.)  Diagrams should be no more than
 *			     70 columns wide, not counting comment characters.  
 *			     Within an arglist width may be 60, and within an 
 *			     intlist, 45.
 *	\intn	\endliteral  The end of the literal  section.   No text 
 *				may be included on this line either.
 *	\intlend
 *\arglend
 *
-*/

/* AUTHOR:
 *
 *	Debra S Baddorf		June 1986
 *
 * PUBLISHED INFORMATION:
 *
 *	document not yet numbered
 *
 * MODIFICATION HISTORY:
 *
 *	V1.0	Deb Baddorf	7/25/86		Initial version released
 *
 * KNOWN BUGS:
 *
 *
 * CHANGE REQUESTS:
 *
 *
 */

/*
 * EQUATED SYMBOLS:
 */

#include	stdio			/* standard I/O parameters */
#include	ctype			/*  ?? what's in here ??   */
#include	ssdef			/* status return symbols   */
#include	curses			/* I use TRUE and FALSE from here */
#include	rmsdef			/* RMS symbol definitions  */
#include	descrip			/* descriptor structures   */
#include	climsgdef		/* CLI (command line interpretter) */
					/* definitions		   */

/*
 * INCLUDED CODE, ETC.
 */

	/* defines known file extensions, and a structure to hold them   */
#include	"extensions.h"
	/* string pattern match routine   */
#include	"match_string.c"
		/* library of useful routines     */
#include	"dsblib.c"

/*
 * MACRO DEFINITIONS:
 */

#define	MIN(x,y)	((x) > (y) ? (y) : (x))
#define	MAX(x,y)	((x) < (y) ? (y) : (x))
#define	ODD(num)	((num) & 1)

/*
 * SYMBOL DEFINITIONS:
 */
/*   	(if defined, EXTRACTED.TMP file not deleted)
 *   #define DEBUG 1       
 */

#define OPEN   1		/* codes for opn_clos_code routine */
#define CLOSE  2
#define FIRST  1		/* codes as to what kind of header line */
#define OTHER  2		/* this is                              */
#define LAST   3
#define TITLE  4
#define HELP   5
#define ALPHA  6

/*
 * GLOBAL DECLARATIONS:
 */

extern int	LIB$SIGNAL();
extern int	LIB$SPAWN();
extern int	LIB$GET_FOREIGN();
extern int	CLI$DCL_PARSE();
extern int	CLI$PRESENT();
extern int	CLI$GET_VALUE();
extern int	LIB$GET_INPUT();
extern int	INPUT_COMMANDS();	/* module created by EXTRACT_CMD.CLD */

struct filename_array {
	char	log_file[30];
	char	list_file[100];
	char	code_file[100];
	char	internal_file[30];
	char	plain_file[100];
	char	help_file[100];
	char	runoff_file[100];
	char	latex_file[100];
    } files;

struct fileptr_array {
	FILE	*log_file;
	FILE	*list_file;
	FILE	*code_file;
	FILE	*internal_file;
	FILE	*plain_file;
	FILE	*help_file;
	FILE	*runoff_file;
	FILE	*latex_file;
    } fp;

struct phrase_pair  begin_block[MAX_DELIMITERS + 1];
struct phrase_pair  end_block[MAX_DELIMITERS + 1];
struct phrase_pair  mid_block[MAX_DELIMITERS + 1];

/*
 * GLOBAL FLAGS:
 */

bool	runoff    = FALSE;
bool	latex     = FALSE;
bool	help      = FALSE;
bool	plain     = FALSE;

bool	remove    = TRUE;
bool	log       = TRUE;
bool	notquiet  = TRUE;
bool	embed     = TRUE;
bool	multiple  = TRUE;
bool	alpha     = TRUE;

bool	level[10] = {FALSE, FALSE, FALSE, FALSE, FALSE, 	/* level[0] not used */
		     FALSE, FALSE, FALSE, FALSE, FALSE};

/*
 * GLOBAL VARIABLES:
 */

int	help_level = 1;
int	file_count = 0;
int	blocks_taken = 0;
int	header_lines = 0;
int	block_count = 0;
bool	in_header = FALSE;
bool	save_header = FALSE;

/*
 *  SYMBOLS and GLOBALS involved in embedded command search
 */

#define	FILENAME	1
#define	EXT_START	2
#define	SUBNAM		3
#define	SUBCAL		4
#define	SUBTXT		5
#define	ARGLIST		6
#define	ARGN		7
#define	ARGLEND		8
#define	INTLIST		9
#define	INTN		10
#define	INTLEND		11
#define	LITERAL		12
#define	ENDLITERAL	13
#define	EXT_STOP	14
#define HELP		15
#define PLAIN		16
#define BLOCKNAME	17

struct phrase_pair  cmd_table[] = 
    {	"\\filename",		FILENAME,
	"\\extractstart",	EXT_START,
	"\\subnam",		SUBNAM,
	"\\subcal",		SUBCAL,
	"\\subtxt",		SUBTXT,
	"\\arglist",		ARGLIST,
	"\\argn",		ARGN,
	"\\arglend",		ARGLEND,
	"\\intlist",		INTLIST,
	"\\intn",		INTN,
	"\\intlend",		INTLEND,
	"\\literal",		LITERAL,
	"\\endliteral",		ENDLITERAL,
	"\\extractstop",	EXT_STOP,
	"\\help",		HELP,
	"\\plain",		PLAIN,
	"\\blockname",		BLOCKNAME,
	"  ",			0
    };

bool	found_cmd[sizeof(cmd_table) / sizeof (struct phrase_pair)];
bool	found_any = FALSE;
char	from_file[100] = "NO FILE NAME";


main ()
{
	char	errmsg[]="Unrecognized extension on file %s\n";
	char	infomsg[]="Working on file %s\n";
	char	cont_msg[] = "Processing embedded commands:\n\n";
	char	alphmsg[] = "Alphabetizing modules.\n\n";
	char	alpherr[]=
		  "Cannot alphabetize due to missing NAME on some modules.\n\n";
	char	blank[]=" ";
	int	file_status=1, errors=0, flag_warnings=0, flag_errors=0;
	int	file_type, status, alph_errs=0;  
	int	i;	/* test  ~~/\~~  */

	strcpy (files.log_file, "EXTRACTED.LOG");		/* fill in file names for set files */
	strcpy (files.internal_file, "EXTRACTED.TMP");

	status = process_input ();
	if (!ODD(status)) exit (0);

	status = open_files();
	if (!ODD(status)) exit (0);

	if (multiple)					/* read a file name into files.code_file */
		file_status = get_next_file ();

	while (file_status != EOF)
	    {	file_type = get_file_type (files.code_file);
		if (file_type < 0 || file_type > ext_count)	/* not recognized type */
		    {	error_msg (errmsg, files.code_file);
			if (multiple)
			    file_status = get_next_file();
			else
			    file_status = EOF;
		    }
		else
		    {	if (opn_clos_code (OPEN))		/* open code_file  */
			    {	file_count++;
				error_msg (infomsg, files.code_file);
								/* process it */
				errors += !process_code (file_type, &alph_errs);
				opn_clos_code (CLOSE);		/* close code_file */
			    }
			if (multiple)			/* if multiple, loop back for more files */
				file_status = get_next_file ();
			else	file_status = EOF;
		    }
	    }		/* while file_status != EOF */
	write_summary (1, file_count, errors);			/* write first summary */
	if (blocks_taken)					/* skip rest of work if no blocks found */
	    {	if (alpha && blocks_taken > 1)
		    {	if (!alph_errs)				/* alphabetize blocks */
			    {	error_msg (alphmsg, blank);
				alphabetize ();
			    }
			else	error_msg (alpherr, blank);
		    }
		if (fseek (fp.internal_file, 0, 0))
		    {	error_msg ("Cannot rewind file %s.\n", files.internal_file);
			exit (0);
		    }
		error_msg (cont_msg, blank);
		process_embedded_cmds (&flag_errors, &flag_warnings);
		write_summary (2, flag_errors, flag_warnings);
	    }
	close_files ();

	/* TEST  ~~/\~~  */
TESTIT: ;
/*	if (multiple)	printf ("Multiple is TRUE\t");
	else		printf ("Multiple is FALSE\t"); 
	printf ("File_type is: %d\n", file_type);
	printf ("Runoff: %d\t\t", runoff);
	printf ("Remove: %d\n", remove);
	printf ("Latex: %d\t\t", latex);
	printf ("Log: %d\n", log);
	printf ("Help: %d\t\t", help);
	printf ("\tNotquiet: %d\n", notquiet);
	printf ("Plain: %d\t\t", plain);
	printf ("Embed: %d\n", embed);
	printf ("Alpha: %d\t\t", alpha);
	printf ("Helplevel: %d\n", help_level);
	printf ("Level: %d  ", level[0]);
		for (i=1; i<10; i++)  printf ("%d  ", level[i]);
		printf ("\n");
	printf ("List_file is: %s\n", files.list_file);
	printf ("Code_file is: %s\n", files.code_file);
	printf ("Plain_file is: %s\n", files.plain_file);
	printf ("Help_file is: %s\n", files.help_file);
	printf ("Latex_file is: %s\n", files.latex_file);
	printf ("Runoff_file is: %s\n", files.runoff_file);
*/
}

process_input ()
{
	char	infile[100]=" ", outfile[100]=" ";
	int	status;

	/* collect input command line */
	status = get_input_commands ();
	if (!ODD(status))  return (0);

	/* set flags and collect values from command line */
	status = process_commands(infile, outfile);
	if (!ODD(status))  return (0);

	/* check infile and outfile names & process them.
	 * put names in proper spaces in FILES structure  */
	status = handle_files (infile, outfile);
	return (status);

}

get_input_commands ()
{
	char	string[246];
	$DESCRIPTOR(str_des, string);
	char	full_command_string[256];
	$DESCRIPTOR(full_cmd_des, full_command_string);
	char	verb[] = "EXTRACTOR ";
	short int	length;
	int	status;

	status = LIB$GET_FOREIGN (&str_des, 0, &length);
	if (!ODD(status))
	    {	/*  LIB$SIGNAL (status);  */
		return (0);
	    }

	strcpy (full_command_string, verb);
	strcpy (full_command_string+10, string);
	status = CLI$DCL_PARSE (&full_cmd_des, INPUT_COMMANDS,
			&LIB$GET_INPUT, &LIB$GET_INPUT);
	if (!ODD(status))
	    {	/*  LIB$SIGNAL (status);  */
		return (0);
	    }
	return (1);
}

/*  PROCESS_COMMANDS examines command string for specific commands, and sets
 *   flags to indicate which were present or defaulted.  Returns 1 if 
 *   successful, 0 otherwise.  Returns an input and output file spec if found.
 */

process_commands(infile, outfile)
char	*infile, *outfile;
{
	$DESCRIPTOR (alpha_descrip,      "ALPHABETIZE");
	$DESCRIPTOR (output_descrip,     "OUTPUT");
	$DESCRIPTOR (level_descrip,      "LEVEL");
	$DESCRIPTOR (hlplev_descrip,     "HELP_LEVEL");
	$DESCRIPTOR (remove_descrip,     "REMOVE");
	$DESCRIPTOR (log_descrip,        "LOG");
	$DESCRIPTOR (quiet_descrip,      "QUIET");
	$DESCRIPTOR (embed_descrip,      "EMBEDDED");
	$DESCRIPTOR (processing_descrip, "PROCESSING");
	$DESCRIPTOR (input_descrip,      "P1");
	FILE	*fopen();
	struct  dsc$descriptor_s    *ptr;
	int	status;
	int	okaysum = 1, ok;		/* If okaysum remains true, then all is ok. */
	char	value_str[80];
	$DESCRIPTOR (val_descrip, value_str);
	short int	len;



	ptr = &quiet_descrip;				/* check QUIET qualifier */
	    status = CLI$PRESENT (ptr);
	    notquiet = (status != CLI$_PRESENT);
	ptr = &log_descrip;				/* check LOG qualifier */
	    status = CLI$PRESENT (ptr);
	    log = (status != CLI$_NEGATED);
	    if (log)					/* open log file */
		{   fp.log_file = fopen (files.log_file, "w");
		    if ((fp.log_file == NULL) && notquiet) 
			{   printf ("Cannot open file %s\n", files.log_file);
			    okaysum = 0;
			}
		}
	ptr = &remove_descrip;				/* check REMOVE qualifier */
	    status = CLI$PRESENT (ptr);
	    remove = (status != CLI$_NEGATED);
	ptr = &embed_descrip;				/* check EMBEDDED qualifier */
	    status = CLI$PRESENT (ptr);
	    embed = (status != CLI$_NEGATED);
	    embed = (embed && remove);			/* if NOREMOVE (remove is false) embed = F */
	ptr = &alpha_descrip;				/* check ALPHABETIZE qualifier */
	    status = CLI$PRESENT (ptr);
	    alpha = (status != CLI$_NEGATED);
	ptr = &level_descrip;				/* check LEVEL qualifier */
	    status = CLI$PRESENT (ptr);
	    {	status = SS$_NORMAL;
		while (status != CLI$_ABSENT)
		    {	status = CLI$GET_VALUE (ptr, &val_descrip, &len);
			if (status != CLI$_ABSENT) 
			    {	ok = get_level (value_str, len);	/* process level phrase */
				okaysum &= ok;
			    }
		    }
	    }
	ptr = &hlplev_descrip;				/* check HELP_LEVEL qualifier */
	    status = CLI$PRESENT (ptr);
	    status = CLI$GET_VALUE (ptr, &val_descrip, &len);
		if (status != CLI$_ABSENT) 
		    {	value_str[len] = '\0';
			sscanf (value_str, "%d", &help_level);	/* get number */
			if (help_level > 9) help_level = 9;
			if (help_level < 1) help_level = 1;
		    }
	ptr = &processing_descrip;			/* check PROCESSING qualifier */
	    status = CLI$PRESENT (ptr);
	    {	status = SS$_NORMAL;
		while (status != CLI$_ABSENT)
		    {	status = CLI$GET_VALUE (ptr, &val_descrip, &len);
			if (status != CLI$_ABSENT) 
			    {	ok = get_type (value_str);	/* get processing type */
				okaysum &= ok;
			    }
		    }
	    }
	ptr = &input_descrip;				/* check INPUT file qualifier */
	    status = CLI$PRESENT (ptr);
	    if (status == CLI$_PRESENT)
		{   status = CLI$GET_VALUE (ptr, &val_descrip, &len);
		    if (status != CLI$_ABSENT) 
			{   *(value_str+len) = '\0';
			    strcpy (infile, value_str);	/* pass phrase out; no checking done */
			}
		}
	ptr = &output_descrip;				/* check OUTPUT file qualifier */
	    status = CLI$PRESENT (ptr);
	    if (status == CLI$_PRESENT)
		{   
		    status = CLI$GET_VALUE (ptr, &val_descrip, &len);
		    if (status != CLI$_ABSENT) 
			{   *(value_str+len) = '\0';
			    strcpy (outfile, value_str);/* pass phrase out; no checking done */
			}
		}
	return (okaysum);
}


get_level (string, len)
char	*string;
short int  len;
/* GET_LEVEL  returns true if okay, false if not.  It sets LEVEL flags
 * (array of 10 of them) true depending on LEVEL phrase passed in as "string".
 * Valid level codes:
 *	1
 *	2
 *	...
 *	9
 *	ALL     ( A,  AL )
 *	nTOm	( nTm )   no spaces
 *	n-m		  no spaces
 */
{
#define OKAY    1
#define ALL_CODE  1
	char	errmsg[] = "Invalid LEVEL argument \\%s\\. \n";
	struct	phrase_pair table[] = 
		{  "ALL", ALL_CODE,
		   " "  ,  0         };

	int	i, j, x, j_offset = 0;


	string[len] = '\0';

	if (match_string(string, table, FALSE) == ALL_CODE)	/* ALL     */
	    {	for (i=1; i<10; i++)			/* level[0] not used */
		    level[i] = TRUE;
		return (OKAY);
	    }
	if (len == 1 && isdigit(string[0]))		/* single digit */
	    {	i = atoi (string);
		if (!i) i = 1;				/* currently disallowing 0 */
		level[i] = TRUE;
		return (OKAY);
	    }
	if (isdigit(string[0]))				/* first char is a digit */
	    {
		if (string[1] == '-' && isdigit(string[2])) /* n-m */
		    {	j_offset = 2;    }
		else if (string[1] == 'T')		/* looks like a TO phrase */
		    {	if (string[2] == 'O' && isdigit(string[3]))	/* nTOm */
			    j_offset = 3;
			else if (isdigit(string[2]))			/* nTm  */
			    j_offset = 2;
		    }
	    }
	if (j_offset)				/* if j_offset != 0, then it matched */
	    {	i = atoi (string);		/* i starts at 0 offset */
		j = atoi (string + j_offset);	/* j starts at j_offset */
		if (0<i && i<10 && 0<j && j<10) /* valid 1 thru 9  */
		    {
			for (x=MIN(i,j);  x<=MAX(i,j);  x++)
			    level[x] = TRUE;
			return (OKAY);
		    }
	    }

/* Error return 		  Fall through here if NOT success  */
	error_msg (errmsg, string);
	return (!OKAY);
}


get_type (string)
char	*string;
/* GET_TYPE sets flags  LATEX, RUNOFF, HELP, PLAIN as needed according to
 * processing type specified on input line.  Returns 1 if okay, 0 if not.
 * Valid types:
 *	PLAIN	print unadulterated extracted block to a file
 *	HELP	print extracted block to a file including help level numbers
 *	LATEX	make substitutions for embedded \ commands.  Output contains
 *			\ commands suitable for LATEXing.
 *		This is the default.
 *	RUNOFF	make substitutions for embedded \ commands.  Output contains
 *			. commands suitable for RUNOFFing.
 *	EPIC	==  PLAIN
 *	CLIB	== (PLAIN,HELP)
 *	YES	== (HELP,RUNOFF,LATEX)	
 *	ALL	== (HELP,RUNOFF,LATEX,PLAIN)
 * Input line accepts a string of types too, PROCESSING=(LATEX,RUNOFF,HELP)
 */
{
#define OKAY   1

	switch (*string)
	    {	case 'E':
		case 'P': plain = TRUE;
			  break;

		case 'L': latex = TRUE;
			  break;

		case 'C': plain = TRUE;
		case 'H': help = TRUE;
			  break;

		case 'A': plain = TRUE;
		case 'Y': help = TRUE;
			  latex = TRUE;
		case 'R': runoff = TRUE;
			  break;

		default	: return (!OKAY);	/* error - shouldn't get here anyway */
	    }
	return (OKAY);				/* success */
}

handle_files (infile, outfile)
char	*infile, *outfile;
/*
 *  Check infile and outfile names.  Fill FILES array as necessary.
 */
{
	int	status;

/*  Check the input file specification. */

	status = handle_infile (infile);

/*  Check the output file specification. */

	status &= handle_outfile (outfile);
	return (status);
}

handle_infile (infile)
char	*infile;
/*  Check infile name - if extension is .LIS or is omitted, file is a list of
 *  other file names.  If input specification is omitted entirely, create a
 *  file containing the filenames in current default directory.  In both cases,
 *  put name of multiple file list in FILES.LIST_FILE.  
 *  If infile name is completely specified, and is not .LIS, put name in 
 *  FILES.CODE_FILE.  Validity of extension is checked later.
 */
{
#define EXACT    TRUE
#define LIS_CODE  1
	struct	phrase_pair table[] = 
		{  "LIS", LIS_CODE,
		   " "  ,  0         };
	char	*strrchr(), *dot, *bracket, *semi, *matchit;
	int	status=1, len;

	if (infile[0] == ' ')				/* no (valid) infile name */
	    {	status &= make_dir (infile);
		multiple = TRUE;
		strcpy (files.list_file, infile);	/* put name in FILES array */
	    }
	else					/* there is a file name here */
	    {	dot = strrchr (infile, '.');
		bracket = strrchr (infile, ']');
		semi = strrchr (infile, ';');
						/* is there an extension? */
		if (dot == NULL ||		/* no dot */
		    (bracket != NULL && dot < bracket))	/* dot was inside a [directory] spec */
						/* therefore, there is no file extension */
		    {	strcat (infile, ".LIS");
			multiple = TRUE;
			strcpy (files.list_file, infile);  /* put name in FILES array */
		    }
		else				/* there is a file extension */
		    {				/* is it .LIS ?  */
			len = (semi==NULL) ? 0 : (semi - dot);	/* len is length of extension */
			matchit = strsave (dot+1, len);		/* copy out just the extension */
			if (match_string(matchit, table, EXACT) == LIS_CODE)
			    {	multiple = TRUE;
				strcpy (files.list_file, infile);  /* put name in FILES array */
			    }
			else			/* regular code file */
			    {	multiple = FALSE;
				strcpy (files.code_file, infile);  /* put name in FILES array as the CODE_FILE */
			    }
		    }
	    }
	return (status);
}

handle_outfile (outfile)
char	*outfile;
/*  Check outfile name - if only one type of output file is supposed to be
 *  produced (depends on "processing" requested) then use file spec as is,
 *  if anything is specified.   If more than one type output file, then
 *  try to use user's choice of outfile with our default extensions.  If
 *  user didn't specify anything, use entire default file names.
 */
{
	char	*strrchr(), *dot, *bracket;
	char	*hlp, hlpwith[] = ".HLPRNO", hlpwithout[] = ".HLP";
	char	*ptr, defaultfile[30] = "EXTRACTED";

	/* if only one type output file needed, and if output file is
	 * specified,  use filename given.
	 */
	if ((plain + help + latex + runoff == 1) && isalnum(outfile[0]))
	    {	if      (plain)  ptr = files.plain_file;
		else if (help)   ptr = files.help_file;
		else if (latex)  ptr = files.latex_file;
		else if (runoff) ptr = files.runoff_file;
		strcpy (ptr, outfile);		/* copy file name into proper slot in FILES array */
	    }
	else
	    {	if (outfile[0] != ' ')		/* file specified, so use it */
		    {				/* look for file extension */
			dot = strrchr (outfile, '.');
			bracket = strrchr (outfile, ']');
			if (dot != NULL &&
			    (bracket == NULL || dot > bracket))	/* there IS an extension */
			    {			/* so remove it */
/*				while (*dot != '\0')
*/				    {	*dot = '\0';
/*					dot++;
*/				    }
			    }
					/* now, set pointer to exntension-less outfile name */
			ptr = outfile;
		    }
		else			/* file not specified.  Use default. */
			ptr = defaultfile;
		/*
		 *  Now, put file name with proper extensions into FILES array
		 *  where needed.
		 */

		hlp = embed ? hlpwith : hlpwithout;	/* help file extension depends on EMBED flag */

		strcpy (files.plain_file,  ptr);
		    strcat (files.plain_file,  ".TXT");
		strcpy (files.help_file ,  ptr);
		    strcat (files.help_file ,  hlp);
		strcpy (files.latex_file,  ptr);
		    strcat (files.latex_file,  ".TEX");
		strcpy (files.runoff_file, ptr);
		    strcat (files.runoff_file, ".RNO");

	    }
	return (1);
}

get_file_type (filename)
char	*filename;
/*  get_file_type compares file extension of 'filename' with known extensions
 *  in extensions.h.  Valid returns are 0 through ext_count.  An error condition
 *  will therefore return -1.
 */
{
#define  EXACT  TRUE
	static struct phrase_pair   EXT[ext_count+1];
	static bool	filled = FALSE;
	char	*strrchr(), *dot, *malloc(), *semi, *partial;
	int	i, typecode;

	if (!filled)	/* fill EXT array with phrase pairs to match ... */
			/* from array of known file extensions.          */
	    {	for (i=0;  i < ext_count;  i++)
		    {	EXT[i].code = i+1;	/* add 1, since 0 tells match_string to stop */
			EXT[i].phrase = extensions[i].ext;
		    }
		EXT[ext_count].code = 0;	/* last one must be 0'ed */
		filled = TRUE;
	    }

	dot = strrchr (filename, '.');
	semi = strrchr (filename, ';');
	if (dot == NULL || (semi != NULL && semi < dot))
		return (-1);			/* no .  or  semi before .  */
	if (semi == NULL)
		i = 0;
	else    i = (int) semi - (int) dot;	/* length of extension (including room for '\0' */
	partial = strsave (dot+1, i);		/* copy out just the extension */

	typecode = match_string (partial, EXT, EXACT);
	if (!typecode || typecode == ambiguous)
		return (-1);			/* -1 signifies error */
	else
		return (typecode - 1);		/* 0 is a valid return */
}

get_next_file ()
/*  Read from file  files.list_file  (already open as fp.list_file). 
 *  Put name into files.code_file.   
 */
{
#define MAXLINE  99
	char	string[MAXLINE+1],  *fgets(), *ptr, *newline,
		blank[]=" ",  *dot, *semi, *br, *strrchr();
	bool	notafile = TRUE;

	while (notafile)
	    {
		ptr = fgets(string, MAXLINE, fp.list_file);
		if (ptr == NULL)
		    {	strcpy (files.code_file, blank);
			return (EOF);
		    }
	    /* Uppercase the whole word.  If file was created by
	     * DIR, it already is uppercase, but if it is a user
	     * supplied file list, it may not be.
	     */
		while (*ptr)
		    {	*ptr = _toupper (*ptr);
			ptr++;
		    }
	    /* Last char is probably a \newline.  Look for it, and
	     * Remove it, or extension match routine has problems.
	     */
		newline = strrchr (string, '\n');
		if (newline != NULL) *newline = '\0';

	    /* Find period (after bracket) and semicolon.  Determine
	     * length of extension.
	     */
		dot = strrchr (string, '.');
		br = strrchr (string, ']');
		semi = strrchr (string, ';');
		if (dot != NULL && (br == NULL || dot > br) &&
			(semi == NULL || semi > dot))
		    {	notafile = FALSE;	/* looks like a file.  Will
						 * let get_file_type decide.
						 */
			strcpy (files.code_file, string);
			return(1);
		    }
	    }		/* end while notafile  */
	strcpy (files.code_file, blank);	/* blank name just in case, */
	return (0);				/* but should never reach here */
}

make_dir (filename)
char	*filename;
/*
 * Make a list of the files in current directory.  Put them in any file you
 * please, but put file name assigned in string "filename".  Return status.
 */
{
	char	cmd[] = 
"DIRECTORY/NOHEADING/NOTRAILING/COLUMNS=1/VERSIONS=1/OUTPUT=EXTRACTED.NAMES";
	$DESCRIPTOR (dir_cmd, cmd);
	unsigned long int status;

	strcpy (filename, "EXTRACTED.NAMES");
	status = LIB$SPAWN (&dir_cmd);
	return (status);
}

open_files ()
/*  Open all files except actual input code file, and log file (which is
 *  already open).
 */
{
	FILE	*fopen();
	char	errmsg[] = "Cannot open file %s\n";
	int	status = 1;

	if (multiple)
	    {	fp.list_file = fopen (files.list_file, "r");
		if (fp.list_file == NULL)
		    {	error_msg (errmsg, files.list_file);
			status = 0;
		    }
	    }
	if (plain)
	    {	fp.plain_file = fopen (files.plain_file, "w");
		if (fp.plain_file == NULL)
		    {	error_msg (errmsg, files.plain_file);
			status = 0;
		    }
	    }
	if (help)
	    {	fp.help_file = fopen (files.help_file, "w");
		if (fp.help_file == NULL)
		    {	error_msg (errmsg, files.help_file);
			status = 0;
		    }
	    }
	if (runoff)
	    {	fp.runoff_file = fopen (files.runoff_file, "w");
		if (fp.runoff_file == NULL)
		    {	error_msg (errmsg, files.runoff_file);
			status = 0;
		    }
	    }
	if (latex)
	    {	fp.latex_file = fopen (files.latex_file, "w");
		if (fp.latex_file == NULL)
		    {	error_msg (errmsg, files.latex_file);
			status = 0;
		    }
	    }
	fp.internal_file = fopen (files.internal_file, "w+");
		if (fp.internal_file == NULL)
		    {	error_msg (errmsg, files.internal_file);
			status = 0;
		    }
	return (status);
}

error_msg (msg, arg)
char	*msg, *arg;
/*  Checks LOG and NOTQUIET flags, and, as appropriate, writes string
 *  msg with argument string arg.   String msg must have embedded printing
 *  format command %s in order to utilize string arg.
 */
{
	if (notquiet) fprintf (stderr, msg, arg);
	if (log)      fprintf (fp.log_file, msg, arg);
}

write_msg_n (msg, arg)
char	*msg;
int	arg;
/*  Checks LOG and NOTQUIET flags, and, as appropriate, writes string
 *  msg with integer argument arg.   String msg must have embedded printing
 *  format command %d in order to utilize integer arg.
 */
{
	if (notquiet) fprintf (stderr, msg, arg);
	if (log)      fprintf (fp.log_file, msg, arg);
}

close_files ()
/*  Close all files except code file.  */
{
	char	errmsg[]="Error closing file %s\n";
	int	status, fclose();
	char	clean_up[160] = "DELETE  ";
	$DESCRIPTOR (clean, clean_up);

	if (multiple)
	    {	status = fclose (fp.list_file);
		if (status) error_msg (errmsg, files.list_file);	/* 0 is success */
	    }

	status = fclose (fp.internal_file);
	if (status) error_msg (errmsg, files.internal_file);

#ifndef DEBUG
	/* delete internal file unless symbol DEBUG is defined  */
	strcat (clean_up, files.internal_file);	/* add correct file name */
	strcat (clean_up, ";0");		/* only top version */
	clean.dsc$w_length = strlen (clean_up);	/* fix string length */
	LIB$SPAWN (&clean);
#endif

	if (plain)
	    {	status = fclose (fp.plain_file);
		if (status) error_msg (errmsg, files.plain_file);
	    }

	if (help)
	    {	status = fclose (fp.help_file);
		if (status) error_msg (errmsg, files.help_file);
	    }

	if (runoff)
	    {	status = fclose (fp.runoff_file);
		if (status) error_msg (errmsg, files.runoff_file);
	    }

	if (latex)
	    {	status = fclose (fp.latex_file);
		if (status) error_msg (errmsg, files.latex_file);
	    }

	if (log)
	    {	status = fclose (fp.log_file);
		if (status) 
		    if (notquiet) fprintf (stderr, errmsg, files.log_file);
	    }
}

opn_clos_code (action)
int	action;
{
	int	status;
	FILE	*fopen();
	char	errmsgO[] = "Error opening code file %s\n";
	char	errmsgC[] = "Error closing code file %s\n";

	switch (action)
	    {	case OPEN  :	strcpy (from_file, files.code_file);
				fp.code_file = fopen (files.code_file, "r");
				if (fp.code_file == NULL)
				    {	error_msg (errmsgO, files.code_file);
					return (0);
				    }
				else	return (1);
		case CLOSE :	status = fclose (fp.code_file);
				if (status)			/* 0 is success */
				    {	error_msg (errmsgC, files.code_file);
					return (0);
				    }
				else	return (1);
	    }
}

process_code (filetype, alph_errs)
int	filetype;
int	*alph_errs;
{
#define  MAXLEN  140		/* allow for embedded control chars */
	char	iline[MAXLEN+1] = "", blank[] = " ";
	char	newline[] = "\n";
	int	source_lines = 0, error_count = 0;

	in_header = save_header = 0;			/* set to 0 for each code file */
	header_lines = block_count = 0;
	fill_match_arrays (filetype);
	while (fgets (iline, MAXLEN, fp.code_file) != NULL)
	    {	source_lines++;
		process_line (iline, filetype, &error_count, alph_errs);
	    }			/* end of file */
	if (in_header)
	    {	error_msg ("  Header unterminated at EOF.\n", blank);
		error_count++;
		if (save_header)
		    store_line (blank, LAST);		/* terminate header block correctly */
	    }
	write_msg_n ("    %d source lines read.  ", source_lines);
	write_msg_n ("%d header lines written.\n", header_lines);
	if (block_count)
		write_msg_n ("    %d header blocks encountered.\n", block_count);
	else
	    {	error_msg ("    >>No documentation blocks in this file.\n", blank);
		error_count++;
	    }
	if (error_count)
		write_msg_n ("    %d errors found.\n", error_count);
	return (error_count == 0);			/* return 1 if no errors */
}

fill_match_arrays (filetype)
int	filetype;
{
	int	i = 0;

	/* fill comment delimiting arrays with appropriate delimiters
	 * for this file type.   
	 */
	for (i=0; i<extensions[filetype].begin_count; i++)
	    {	begin_block[i].phrase = extensions[filetype].begin[i].string;
		begin_block[i].code = strlen(begin_block[i].phrase) + 1;
	    }

	for (i=0; i<extensions[filetype].end_count; i++)
	    {	end_block[i].phrase = extensions[filetype].end[i].string;
		end_block[i].code = strlen(end_block[i].phrase) + 1;
	    }

	for (i=0; i<extensions[filetype].mid_count; i++)
	    {	mid_block[i].phrase = extensions[filetype].mid[i].string;
		mid_block[i].code = strlen(mid_block[i].phrase) + 1;
	    }
}

get_line_type (line, comment_len)
char	*line;
int	*comment_len;
/*  Decide whether line is the FIRST in a header block (has a + in the right
 *  place), the LAST (has a -), or OTHER.  Returns the length of the comment
 *  string detected at beginning of line, if line matched one of the known
 *  comment patterns.
 */
{
#define  NOEXACT  FALSE
	int	i;
	bool	match = FALSE;

	i = match_string (line, begin_block, NOEXACT);
	if (match = (  i   && i != ambiguous))			/* is it a FIRST? */
	    {	*comment_len = i-1;
		return (FIRST);
	    }
	if (!match)
	    {	i = match_string (line, end_block, NOEXACT);
		if (match = (  i   && i != ambiguous))		/* is it a LAST? */
		    {	*comment_len = i-1;
			return (LAST);
		    }
	    }
	if (!match)
	    {	i = match_string (line, mid_block, NOEXACT);
		if (match = (  i   && i != ambiguous))		/* is it even a comment? */
		    {	*comment_len = i-1;
			return (OTHER);
		    }
	    }
	return (OTHER);						/* must be a line of code */
}

store_line (line, linetype)
char	*line;
int	linetype;
/*  Store line in header file.  Pointer must point after comment delimiter,
 *  if these are to be stripped.  This routine just prints what it is passed.
 *  Some needed embedded commands are added, depending on the linetype.
 */
{
	char	firstline[] = "\\extractstart\n";
	char	endline[] = "\\extractstop\n";

	switch (linetype)
	    {	case FIRST : if (embed)
			    	  fprintf (fp.internal_file, "%s", firstline);
		case OTHER : fprintf (fp.internal_file, "%s", line);
			     header_lines++;
			     break;
		case LAST  : fprintf (fp.internal_file, "%s", line);
			     if (embed)
				  fprintf (fp.internal_file, "%s", endline);
			     break;
		case TITLE :
		case ALPHA :
		case HELP  : fprintf (fp.internal_file, "%s\n", line);
	    }
}

hdr_level_test (line, errptr)
char	*line;
int	*errptr;
/*  Line must point at the LEVEL character after the + on header start line.
 */
{
	int	i = 1;
	char	c[2] = " ";

	if (isdigit (*line))
	    {	*c = *line;		/* get single char, in case user has typed > 1 */
		i = atoi (c);
		return (i);
	    }
	else if (*line == ' ')
		return (1);
	else
	    {	/* default all other cases to 0. These cases
		 * will NOT get printed, since level[0] is FALSE 
		 * and is not used.
		 */
		(*errptr)++;
		error_msg ("  Invalid LEVEL character: %s", line);	/* line has a \newline char in it */
		return (0);
	    }
}

process_line (line, filetype, errptr, alpherr)
char	*line;
int	filetype, *errptr, *alpherr;
{
#define SYMBOL_LEN  2		/* length of "+n" in  "/*+nNAME"  */
	int	linetype, comment_len = 0, offset = 0, hdr_level;
	int	ok;
	static char	new_name[50] = "", old_name[50];
	char	extract[] = "  Extracting header %s.\n";
	char	skip[] = "  Skipping header %s.\n";
	char	errmsg[] = "  Header %s started ";
	char	errmsgcnt[] = "without ending header %s.\n";
	char	namemsg[] = 
		    "  Header has no name.  Cannot do /HELP or /ALPHABETIZE.\n";
	char	help_line[60] = "\\help   ";
	char	alph_line[60] = "1 ";
	char	title_line[240] = "\\filename ";
	char	block_line[120] = "\n\\blockname ";
	char	plain_line[150] = "\\plain ";

	linetype = get_line_type (line, &comment_len);	/* comment_len INCLUDES the + or - */
	switch (linetype)
	    {	case OTHER :
			if (save_header)
			    {	if (remove)  offset = comment_len;
				store_line (line+offset, OTHER);
			    }
			break;
		case FIRST :
			strcpy (old_name, new_name);		/* save current name as old */
			hdr_level = hdr_level_test (line + comment_len, errptr);
			ok = get_name (line + comment_len + 1, new_name);
			if ((ok <= 0) && (help || alpha))
			    {	error_msg (namemsg, new_name);
				(*alpherr)++;
				(*errptr)++;
			    }
			if (in_header)
			    {	error_msg (errmsg, new_name);
				error_msg (errmsgcnt, old_name);
				(*errptr)++;
				if (save_header) store_line (" ", LAST);  /* terminate header block correctly */
			    }
			in_header = TRUE;
			save_header = level[hdr_level];
			block_count++;
			if (!save_header)
			    {	error_msg (skip, new_name); /* msg for skipping */
				break;		/* stop here if not saving block */
			    }
			blocks_taken++;			/* count it, if saving it */
			error_msg (extract, new_name);  /* msg for extracting */
			if (alpha)			/* store line to alphabetize on */
			    {	strcat (alph_line, new_name);
				store_line (alph_line, ALPHA);
			    }
			strcat (title_line, from_file);	/* store filename */
			strcat (block_line, new_name);	/* store blockname */
			strcat (title_line, block_line);/* do both with 1 'store' */
			store_line (title_line, TITLE);
			if (plain)			/* store actual line */
			    {	if (remove)  offset = comment_len + 1;
				strcat (plain_line, line+offset);
				store_line (plain_line, FIRST);
			    }
			if (help)			/* store help key */
			    {	help_line[6] = '0' + help_level;  /* insert # */
				strcat (help_line, new_name);
				store_line (help_line, HELP);
			    }
			break;		/* end case FIRST */
		case LAST  :
			if (!in_header)
			    {	error_msg ("  Loose header ending found.\n", skip);
				(*errptr)++;
			    }
			else
			    {	if (save_header)
				    {	if (remove)  offset = comment_len;
					strcat (plain_line, line+offset);
					store_line (plain_line, LAST);
				    }
				in_header = save_header = FALSE;
			    }
			break;
	    }		/* end switch */
}

write_summary (code, num1, num2)
int	code, num1, num2;
{
	char	blank[] = " ";
	char	long_msg[] = 
		   "and %d warnings while processing embedded commands.\n";
	char	fix_msg[] = "Please fix the errors before proceeding.\n\n";

	switch (code)
	    {	case 1:	write_msg_n ("Of %d files, ", num1);
			write_msg_n ("%d had errors.\n\n", num2);
			break;
		case 2: write_msg_n ("Found %d errors, ", num1);
			write_msg_n (long_msg, num2);
			if (num1)
				write_msg_n (fix_msg, num1);
	    }
}

get_name (line, name)
char	*line, *name;
/*  Line must point to start of name  */
{
	int	status;

	status = sscanf (line, "%s", name);
	return (status);
}

process_embedded_cmds (err, warn)
int	*err, *warn;
{
#define	MAX_LEN  200
	char	line[MAX_LEN+1], *char_ptr;
	char	*fgets(), *handle_cmd();

	write_headers();
					/* while there IS a next line */
	while (fgets (line, MAX_LEN, fp.internal_file) != (char *) NULL)
	    {
		char_ptr = line;	/* point to start of line */

					/* check first for alphabetize key,
					 * 1 NAME
					 * which must be removed
					 */
		if (alpha && *char_ptr == '1' &&
		    *(char_ptr+1) == ' ' &&
		    strlen (char_ptr) > 2)
			continue;	/* skip printing this line */

					/* if in a LITERAL block, preserve
					 * white space; else, swallow up
					 * space to start of text or command.
					 */
		while (*char_ptr && isspace (*char_ptr))
			if (found_cmd[LITERAL])
				handle_space (char_ptr++);
			else
			    {	write_plain ("%.1s", char_ptr);	/* DO copy space to plain file */
				if (!embed) write_help ("%.1s", char_ptr); /* ditto to noembed help file */
				char_ptr++;
			    }

		if (*char_ptr == '\0') continue;	/* get next line */
		if (*char_ptr == '\\')
			char_ptr = handle_cmd (char_ptr, err, warn);
					/* function returns pointing either:
					 *    after the command word(s) or
					 *    still at the \ if it is not a
					 *        command.
					 */
		if (*char_ptr)
			handle_rest (char_ptr);
					/* copy out rest of characters, making
					 * substitutions as needed for unprint-
					 * able characters in latex
					 */
	    }			/* end of while.  Go get another line */
	write_footers ();

}

write_headers ()
{
	char	RNO[] = 
".REQUIRE \"EXTRACTOR_INC:EXT_RNO_HDR.INC\"  \n";
	char	HLP[] = 
".REQUIRE \"EXTRACTOR_INC:EXT_HLP_HDR.INC\"  \n";
	char	TEX[] = 
"\\input {EXTRACTOR_INC:EXT_LATEX_HDR.INC}  \n";

	if (embed && help)
		fprintf (fp.help_file, HLP);
	if (runoff)
		fprintf (fp.runoff_file, RNO);
	if (latex)
		fprintf (fp.latex_file, TEX);
}



write_footers ()
{
	char	TEX[] = "\\end{document}  \n";

	if (latex)
		fprintf (fp.latex_file, TEX);
}

handle_space (cptr)
char	*cptr;
{
	char	achar[] = "%.1s",  aword[] = "%s";

	switch (*cptr)
	    {	case ' '  :			/* blank           */
		case '\n' : 			/* newline         */
			    write_it (achar, cptr);
			    break;
		case '\v' :			/* vertical tab    */
		case '\r' :			/* carriage return */
		case '\f' : 			/* form feed       */
			    write_it (achar, "\n");
			    break;
		case '\t' :			/* tab             */
			    write_tex (achar, cptr);
			    write_plain (achar, cptr);
			    write_rno (aword, "    ");
			    if (embed)
				write_help (aword, "    ");
			    else 
				write_help (achar, cptr);
			    break;
	    }
}

write_it (fmt, ptr)
char	*fmt, *ptr;
{
	write_rno_hlp (fmt, ptr);
	write_tex (fmt, ptr);
	write_plain (fmt, ptr);
}


write_tex (fmt, ptr)
char	*fmt, *ptr;
{
	if (latex)
		fprintf (fp.latex_file, fmt, ptr);
}


write_rno (fmt, ptr)
char	*fmt, *ptr;
{
	if (runoff)
		fprintf (fp.runoff_file, fmt, ptr);
}


write_help (fmt, ptr)
char	*fmt, *ptr;
{
	if (help)
		fprintf (fp.help_file, fmt, ptr);
}


write_rno_hlp (fmt, ptr)
char	*fmt, *ptr;
{		/* only called if (embed) -- meaning help gets same as rno file */
	if (runoff)
		fprintf (fp.runoff_file, fmt, ptr);
	if (help)
		fprintf (fp.help_file, fmt, ptr);
}


write_plain (fmt, ptr)
char	*fmt, *ptr;
{
	if (plain)
		fprintf (fp.plain_file, fmt, ptr);
}

handle_rest (ptr)
char	*ptr;
{
	char	achar[] = "%.1s", aword[] = "%s";

	while (*ptr)			/* stop at the '\0' */
	    {	if (isalnum (*ptr))
			write_it (achar, ptr);
		else if (isspace (*ptr))
			handle_space (ptr);
		else if (!found_cmd[LITERAL])	/* no subs inside a literal blk */
		    switch (*ptr)
		    {	case '_' : write_rno (achar, "_");	/* need __ */
			case '&' :
			case '$' :
			case '%' :
			case '#' :
			case '{' :
			case '}' : write_tex (achar, "\\");
				   write_it (achar, ptr);
				   break;
			case '~' : write_tex (aword, "\\verb@");
				   write_it (achar, ptr);
				   write_tex (achar, "@");
				   break;
			case '\\': write_tex (aword, "$\\backslash$");
				   write_rno_hlp (achar, ptr);
				   write_plain (achar, ptr);
				   break;
			case '^' : write_tex (aword, "$\\wedge$");
				   write_rno_hlp (achar, ptr);
				   write_plain (achar, ptr);
				   break;
			case '.' : write_rno (achar, "_");
				   if (embed) write_help (achar, "_");
				   write_it (achar, ptr);
				   break;
			default  : write_it (achar, ptr);
		    }
		else write_it (achar, ptr);
		ptr++;
	    }				/* end while not '\0' */
}

char *
get_next_word (start, copy)
char	*start, *copy;
{
	char	*ptr = start;

	/* put a copy of word into 'copy' */
	sscanf (start, "%s", copy);

	/* point ptr to start of this word in original string */
	/* i.e. move past any white space */
	while (isspace (*ptr))
		ptr++;

	return (ptr);
}

char *
two_word_cmd (cmd, start, orig)
int	cmd;
char	*start, *orig;
{
#define	MAX_LEN  140
	char	line[MAX_LEN+1];
	bool	save_p, save_h;
	char	word2[81]="\0", *ptr, fmt[]="%s", 
		word3[81]="\0", *swallow;
	char	*get_next_word(), *fgets();

NEXT_WD_LOOP:
	ptr = get_next_word (start, word2);	/* ptr now at 1st char of 2nd word */
	if (!word2[0])				/* next word not on this line */
	    {	write_plain ("%s", orig);	/* write out line so far */
					/* if there IS a next line, get it */
		if (fgets (line, MAX_LEN, fp.internal_file) != (char *) NULL)
		    {	strcpy (orig, line);	/* put in original buffer */
			start = orig;		/* start next_word search there */
			goto NEXT_WD_LOOP;
		    }
			/* if no next line, finish up anyway */
	    }
		

	switch (cmd)
	    {	case SUBNAM: write_tex (fmt, "\\subnam{");
			     write_rno (fmt, ".SKIP 1 \n");
			     break;
		case ARGN  : write_tex (fmt, "\\argn{");
			     write_rno_hlp (fmt, ".SKIP 1 \n.INDENT -10;");
			     break;
		case INTN  : write_tex (fmt, "\\intn{");
			     write_rno_hlp (fmt, ".INDENT -10;");
			     break;
	    }
	save_p = plain;				/* don't let handle_rest write */
	plain = FALSE;				/* to plain.  Copied later. */
	if (cmd == SUBNAM)
	    {	save_h = help;			/* don't write SUBNAM to help */
		help = FALSE;			/* file.  Redundant. */
	    }
	handle_rest (word2);			/* Copy word2 into files,    
						 * taking care of any special
						 * characters
						 */
	plain = save_p;
	help = save_h;

	switch (cmd)
	    {	case SUBNAM: write_tex (fmt, "} ");
			     break;
		case ARGN  : write_tex (fmt, "} \\argtxt ");
			     break;
		case INTN  : write_tex (fmt, "} \\inttxt ");
			     break;
	    }
	write_rno_hlp (fmt, "\t");	/* The one place where runoff DOES
					 * need a tab
					 */
	ptr = ptr + strlen(word2);	/* Point at 1st char after second
					 * word - blank or \n probably.
					 */
	swallow = get_next_word (ptr, word3);
					/* Point at 1st char of third word. */
	if (word3[0])
		ptr = swallow;		/* If there IS a third word, point to
					 * that.  We have swallowed up any
					 * space after {word2} of our two-word-
					 * command.  If it doesn't have a
					 * third word, preserve space - and
					 * particularly '\n'.
					 */
	return (ptr);
}

check_case_stop (err, warn)
int	*err, *warn;
{
	char	*ptr, phrase1[] = "\\subcal", phrase2[] = "\\subtxt";
	char	dummy[] = " ";
	char	errmsg[] = "  Error!  File %s ";
	char	wrnmsg[] = "  Warning!  File %s ";
	char	msg1[] = "has a block with no embedded commands.\n";
	char	msg2[] = "has no command %s.\n";
	char	msg3[] = "has mismatched list flag %s.\n";

	if (!found_any)				/* no embedded commands     */
	    {	(*err)++;
		error_msg (errmsg, from_file);
		error_msg (msg1, dummy);
	    }
	else
	    {	if (!found_cmd[SUBNAM])
		    {	(*err)++;
			error_msg (errmsg, from_file);
			error_msg (msg2, "\\subnam");
		    }
		if (!(found_cmd[SUBCAL] == found_cmd[SUBTXT]))
		    {	(*err)++;
			if (!found_cmd[SUBCAL])
				ptr = phrase1;
			else	ptr = phrase2;
			error_msg (errmsg, from_file);
			error_msg (msg2, ptr);
		    }
		if ( found_cmd[ARGLIST] != found_cmd[ARGLEND])
		    {	(*warn)++;			/* END is never set, so it */
			error_msg (wrnmsg, from_file);	/* must be an unpaired LIST */
			error_msg (msg3, "\\arglist");
			write_tex ("%s", "\\arglend\n");
		    }
		if ( found_cmd[INTLIST] != found_cmd[INTLEND])
		    {	(*warn)++;			/* END is never set, so it */
			error_msg (wrnmsg, from_file);	/* must be an unpaired LIST */
			error_msg (msg3, "\\intlist");
			write_tex ("%s", "\\intlend\n");
			write_rno_hlp ("%s", ".LM 15 \n");
		    }
	    }
	write_rno_hlp ("%s", ".LM -5 \n");
	write_tex ("%s", "\\extractstop \n");	/* after handling this command
						 * don't forget to write it to
						 * the file!
						 */
}

check_dups (cmd, count, msg1, msg2, phrase)
int	cmd, *count;
char	*msg1, *msg2, *phrase;
{
	if (found_cmd[cmd])			/* duplicate instance */
	    {	(*count)++;			/* increment ERR or WARN */
		error_msg (msg1, from_file);	/* write messages */
		error_msg (msg2, phrase);
	    }
	found_cmd[cmd]++;			/* set flag TRUE */
}


check_pair (cmd, count, msg1, msg2, phrase)
int	cmd, *count;
char	*msg1, *msg2, *phrase;
{
	if (!found_cmd[cmd])			/* is the paired cmd set? */
	    {	(*count)++;			/* increment ERR or WARN */
		error_msg (msg1, from_file);	/* write messages */
		error_msg (msg2, phrase);
	    }
	found_cmd[cmd] = FALSE;		/* Supposed to go in matching pairs.
					 * Entry to this routine means second
					 * was found (never set TRUE) so set
					 * first FALSE also.
					 */
}

char *
handle_cmd (start, err, warn)
char	*start;				/* points to a leading '\' */
int	*err, *warn;
{
	char	word[81], *ptr = word, word2[81];
	int	cmd, cmd_len, i;
	bool	save_l, save_r, save_p;
	char	save;
	char	from_block[100] = " ";
	char	wrnmsg[] = "  Warning!  File %s ", 
		errmsg[] = "  Error!  File %s ",
		infomsg[] = "Working on header %s, ",
		infomsg2[]= "from file %s.\n",
		msg1[] = "has a misplaced %s.\n",
		msg2[] = "has mismatched list flag %s.\n",
		msg3[] = "  File %s - in \\literal block: ",
		msg4[] = "ignoring embedded command %s.\n",
		fmt[] = "%s";
	char	*get_next_word(), *two_word_cmd();

	sscanf (start, "%80s ", word);		/* copy out the word */
	while (*(++ptr))			/* ignore 1st char ('\') */
		*ptr = _tolower(*ptr);		/* lower case the word */
	cmd = match_string (word, cmd_table, FALSE);
	if (!cmd || cmd==ambiguous)  return (start);
	cmd_len = strlen (word);
	if (embed && found_cmd[LITERAL] && cmd != ENDLITERAL && cmd != EXT_STOP)
					/* ignore cmds in literal block */
	    {	error_msg (msg3, from_file);
		error_msg (msg4, word);
		return (start+cmd_len);
	    }
	found_any++;
	if (!embed && cmd != FILENAME && cmd != BLOCKNAME
		&& cmd != HELP && cmd != PLAIN )
		return (start);		/* only these cmds are valid if not embed */
	switch (cmd)
	    {	case FILENAME : ptr = get_next_word (start+cmd_len, from_file);
				for (i=0; i<sizeof found_cmd; i++)
					found_cmd[i] = FALSE;
				found_any = FALSE;
				*ptr = '\0';		/* don't write extra newline to outputs */
				return (ptr);
		case BLOCKNAME: ptr = get_next_word (start+cmd_len, from_block);
				found_any--;		/* pgm inserted this, so don't count it */
				error_msg (infomsg, from_block); /* let user know something is happending */
				error_msg (infomsg2, from_file);
				*ptr = '\0';		/* don't write extra newline to outputs */
				return (ptr);
		case EXT_START: found_any--;		/* pgm inserted this, so don't count it */
				check_dups (cmd, warn, wrnmsg, msg1, word);
				write_tex (fmt, word);
				write_rno_hlp (fmt, ".SKIP 2 ");
				return (start+cmd_len);
		case EXT_STOP : found_any--;		/* pgm inserted this, so don't count it */
				check_case_stop (err, warn);
				for (i=0; i<sizeof found_cmd; i++)
					found_cmd[i] = FALSE;
				return (start+cmd_len);
		case SUBNAM   : check_dups (cmd, err, errmsg, msg1, word);
				ptr = two_word_cmd (cmd, start+cmd_len, start);
				break;
		case SUBCAL   : check_dups (cmd, err, errmsg, msg1, word);
				write_tex (fmt, "\\subcal  ");
				write_rno (fmt, ".SKIP 1\n");
				write_rno_hlp (fmt, ".LM +15\n.INDENT -10;");
				ptr = start+cmd_len;
				break;
		case SUBTXT   : check_pair (SUBCAL, err, errmsg, msg1, word);
				write_tex (fmt, "\\subtxt ");
				write_rno_hlp (fmt, ".SKIP 1\n.LM -10\n");
				ptr = start+cmd_len;
				break;
		case ARGLIST  : check_dups (cmd, err, errmsg, msg1, word);
				write_tex (fmt, "\\arglist ");
				write_rno_hlp (fmt, ".LM +10 ");
				ptr = start+cmd_len;
				break;
		case ARGN     : if (!found_cmd[ARGLIST])
				    {	(*err)++;
					error_msg (errmsg, from_file);
					error_msg (msg1, word);
				    }
				ptr = two_word_cmd (cmd, start+cmd_len, start);
				break;
		case ARGLEND  : check_pair (ARGLIST, err, errmsg, msg2, word);
				write_tex (fmt, "\\arglend ");
				write_rno_hlp (fmt, ".LM -10 ");
				ptr = start+cmd_len;
				break;
		case INTLIST  : check_dups (cmd, err, errmsg, msg1, word);
				write_tex (fmt, "\\intlist ");
				write_rno_hlp (fmt, ".LM +15 ");
				ptr = start+cmd_len;
				break;
		case INTN     : if (!found_cmd[INTLIST])
				    {	(*err)++;
					error_msg (errmsg, from_file);
					error_msg (msg1, word);
				    }
				ptr = two_word_cmd (cmd, start+cmd_len, start);
				break;
		case INTLEND  : check_pair (INTLIST, err, errmsg, msg2, word);
				write_tex (fmt, "\\intlend ");
				write_rno_hlp (fmt, ".LM -15 ");
				ptr = start+cmd_len;
				break;
		case LITERAL  : check_dups (cmd, err, errmsg, msg1, word);
				write_tex (fmt, "\\literal ");
				write_rno_hlp (fmt, ".LITERAL \n");
				ptr = start+cmd_len;
				break;
		case ENDLITERAL:check_pair (LITERAL, err, errmsg, msg1, word);
				write_tex (fmt, "\\end{verbatim}\\normalsize ");
				write_rno_hlp (fmt, "\n.END LITERAL ");
				ptr = start+cmd_len;
				break;
		case PLAIN    : found_any--;		/* don't count this */
				ptr = start + cmd_len + 1;
				write_plain (fmt, ptr);
				*ptr = '\0';
				return (ptr);
		case HELP     : found_any--;		/* don't count this */
				ptr = start + cmd_len + 1;
				save_l = latex;		/* whata kluge. */
				save_r = runoff;
				save_p = plain;
				latex = runoff = plain = FALSE;
				handle_rest (ptr);
				latex = save_l;
				runoff = save_r;
				plain = save_p;
				*ptr = '\0';
				return (ptr);
	    }		/* end switch */

	save = *ptr;			/* save current character */
	*ptr = '\0';			/* replace it with null */
	write_plain (fmt, start);	/* write line (so far) to plain file */
	*ptr = save;			/* restore original character */
	return (ptr);			/* return; continue processing line */
}

alphabetize ()
{	
	char	*name_ptr;
	FILE	*fopen();
	char	errmsg1[]="  Error closing file %s\n";
	char	errmsg2[] = "  Cannot reopen file %s\n";
	int	status, fclose();

				/* close file first */
	status = fclose (fp.internal_file);
	if (status) error_msg (errmsg1, files.internal_file);	/* 0 is success */

	name_ptr = files.internal_file;	/* get name of file in question */
	alph_kluge (name_ptr);		/* alphabetize it */
				/* reopen file again */
	fp.internal_file = fopen (files.internal_file, "r+");
	if (fp.internal_file == NULL)
		error_msg (errmsg2, files.internal_file);
}

alph_kluge (filename)
char	*filename;	/* easiest way to alphabetize blocks of text is to use
			 * facility designed to handle whole blocks of text -
			 * the LIBRARIAN.  So we "kluge", if you will.
			 * Command formats:
			 * LIBRARY/CREATE/HELP/INSERT  libname  textfile
			 * LIBRARY/HELP/EXTRACT=* /OUTPUT=textfile  libname
			 */

{
	int	status;
	char	in_cmd[] =  "LIBRARY/CREATE/HELP/INSERT ";
	char	out_cmd[] = "LIBRARY/HELP/EXTRACT=*/OUTPUT=";
	char	clean_up[] = "DELETE  _TMPEXTLIB.HLB;0";	/* lib name in only */
	char	libname[] = " _TMPEXTLIB.HLB ";			/* these two places */
	char	string[300] = " ";
	$DESCRIPTOR (cmd, string);
	$DESCRIPTOR (clean, clean_up);
	char	clean_up2[160] = "DELETE  ";
	$DESCRIPTOR (clean2, clean_up2);

		/* build LIB/CREATE command first, with correct file names */
	strcat (string, in_cmd);		/* move command to string */
	strcat (string, libname);		/* append library name */
	strcat (string, filename);		/* add correct file name */
	cmd.dsc$w_length = strlen (string);	/* fix string length */

	status = LIB$SPAWN (&cmd);	/* put into a help library */
	if (!ODD(status)) exit(status);

		/* build LIB/EXTRACT command with correct file names */
	string[0] = '\0';
	strcat (string, out_cmd);		/* move command to string */
	strcat (string, filename);		/* add correct file name */
	strcat (string, libname);		/* append library name */
	cmd.dsc$w_length = strlen (string);	/* fix string length */

	status = LIB$SPAWN (&cmd);	/* pull back out in order! */
	if (!ODD(status)) exit(status);

	status = LIB$SPAWN (&clean);	/* delete library file we've made */

#ifndef DEBUG
	/* delete old version of internal file unless 
	 * symbol DEBUG is defined 
	 */
	strcat (clean_up2, files.internal_file);/* add correct file name */
	strcat (clean_up2, ";-1");	/* previous version, now 
					 * superceded by version in 
					 * alphabetical order
					 */
	clean.dsc$w_length = strlen (clean_up2);/* fix string length */
	LIB$SPAWN (&clean2);
#endif
}
