#module  CC_VPRINTF  "V1.2"

/*|++CCVPRINTF.C

 Facility: 
	Fermilab Accelerator Control System.
	General VAX-11 C Library Routine.

 Abstract:
	This routine provides for non-stream output from VAX-11 C programs to
	SYS$OUTPUT: by using the VAX-11 C library routine sprintf to format a
	string and LIB$PUT_OUTPUT to output the string.

 Environment:
	User mode, linked from FERMILIB into a user's program.
	Program must be linked with SYS$LIBRARY:CRTLIB.OLB.
|--*/

/*
 Modification History:
 Author: F. Nagy			Creation date: 02-Sep-83

 V1.0	02-Sep-83  FJN	Created for use in ACNAUX and ACNETBOOT.
 V1.1	14-Aug-84  FJN	Eliminated LIB_NARGS calls via tricks with pointers,
			did same to giant case to copy arguments
 V1.2	29-Dec-84  FJN	Replace FAIL_TEST macro by ABORT_ON_FAILURE
*/

/* Include Files: */
#include  vaxtypes			/* VAX-11 C Type extensions */
#include  descrip			/* VAX-11 Descriptors */
#include  descrip2			/* Additional descriptor defs. */

/* Local definitions and types: */
typedef unsigned long status;		/* Condition code status type */

#define  SIZE		255		/* Maximum text string size */
#define  MAX_ARGS	20		/* Maximum number of output arguments */

globalvalue STR$_TRU;			/* String truncation error signal */

/* Module local own storage */

static char buffer[SIZE];		/* Text buffer */
static $DESCRIPTOR( d_buffer, buffer);	/* Descriptor of text buffer */
static uword length = 0;		/* Current text length */


/*+0CC_VPRINTF

 Functional Description:
	This routine provides an interface between VAX-11 C and the VMSRTL
	routine LIB$PUT_OUTPUT to write output to SYS$OUTPUT but not as a
	byte stream file.  VAX-11 C formats are used by internally invoking
	the VAX-11 C library routine sprintf.

 Input Parameters:
	format - character string containing the format specifications.
	out<n> - from 0 to 20 output parameters to be encoded into the output
		 text as determined by the format specifications.

 Implicit Inputs:
	NONE

 Output Parameters:
	NONE

 Implicit Outputs:
	NONE

 Condition Codes:
	STR$_TRU	internal buffer overflowed.
	All LIB$PUT_OUTPUT errors are signalled.

 Side Effects:
	Not re-entrant since text string is accumulated until newline
	character is seen then a single output is done.
-*/
/*+ CC_VPRINTF
 Provide simple interface for VAX-11 C to output non-stream file text
 to SYS$OUTPUT via the LIB$PUT_OUTPUT routine and handling standard
 VAX-11 C format specifications (as used by the printf, fprintf and
 sprintf library functions).

   CC_VPRINTF( format.rt.ra, [out1.rz.v], ... [out20.rz.v] )

 format   character string containing the format specification.  In
          VAX-11 C terms this is defined as "char *format".  Passed
          by reference.

 out<n>   From 0 to 20 optional data items or pointers to data items
          referenced by the format specifications for conversion to
          the output text.  Each is treated as an unsigned longword
          and may be a data value passed by value or a pointer to
          data passed by reference.

 The sprintf routine is used to process the format specification and
 LIB$PUT_OUTPUT is used to output the resulting text string.  All
 errors are signalled.  This routine is not re-entrant as the text
 buffer is added to until the final character added by sprintf was
 a newline, then the buffer is output as a single record.
-*/

cc_vprintf( format, o0)
char *format;				/* Format specification text */
ulong o0;				/* "First" output data/pointer */
{
int i,n;
int sprintf();				/* Formatted print to string */
int LIB$CALLG();			/* Indirect call roine */
status LIB$PUT_OUTPUT();		/* Routine to write to SYS$OUTPUT */
status sts;				/* Completion status code */
struct {				/* Argument list for sprintf call */
    int nargs;				/* Number of arguments */
    char *bfr;				/* Pointer to output text buffer */
    char *fmt;				/* Pointer to format string */
    ulong p[MAX_ARGS];			/* output data/pointers */
    } arglist;
ulong *ap;				/* Argument list pointer */

/*
  Get number of arguments in our call and setup fixed portion of the sprintf
  argument list.
*/
n = *( (long *) &format - 1);		/* Get number of call arguments */
arglist.nargs = n + 1;			/* Add buffer argument */
arglist.bfr = &buffer[length];		/* Store at current buffer beginning */
arglist.fmt = format;

/*
  Now copy the output data/pointer arguments into the argument list for the
  sprintf call.  That is, if there are any such arguments.
*/
if (n > 1)
    {
    ap = &o0;				/* Points to first argument */

    for (i = 0; i < (n-1 > MAX_ARGS ? MAX_ARGS : n-1); i++)
	arglist.p[i] = *ap++;		/* Copies actual argument! */
    }

/*
  Now call sprintf in an indirect fashion.  If buffer is overflowed, signal
  an error.
*/
n = LIB$CALLG( &arglist, sprintf);
if (n < 0)
    LIB$STOP( STR$_TRU);

/*
  Now accumulate buffer size and check for terminating character is a newline.
*/
length += n;
if (length == 0)
    return;				/* Buffer is empty */
if (buffer[length-1] != '\n')
    return;				/* Not end of line yet */

/*
  Last character was a new line, output the entire text buffer except for the
  final newline.
*/
d_buffer.dsc$w_length = length - 1;
sts = LIB$PUT_OUTPUT( &d_buffer);
ABORT_ON_FAILURE(sts);

/*
  Reset to empty buffer now.
*/
length = 0;

}
