#module  LIB_TRANSLATE_LOGICAL_NAME  "V1.0"

/* LIBTRNLNM.C

 Facility: 
	Fermilab Accelerator Control System - VAX/VMS ACNET
	General library routine

 Abstract:
	This routine provides a simple interface to the $TRNLNM system service.

 Environment:
	Called from user mode usually.
*/
/*
 Author: Frank J. Nagy		Fermilab Accelerator Controls
 Modification History:

 V0.0	08-Aug-86  FJN	Created to match LIB_CRELNM and for use in DBM
 V1.0	10-Aug-86  FJN	Debugged and inserted into library
*/

/******************
 * Include Files: *
 ******************/
#include  vaxtypes			/* VAX-11 C Type extensions */
#include  descrip			/* VAX-11 Descriptors */
#include  descrip2			/* Additional descriptor defs. */
#include  lnmdef			/* Logical Name definitions */

/********************************
 * Local definitions and types: *
 ********************************/
typedef unsigned long status;		/* Condition code status type */

typedef struct dsc$descriptor *dscptr;	/* Pointer to descriptor */

struct an_item {			/* Element of an item list */
    uword buflen;			/* Item buffer length */
    uword itmcod;			/* Item code */
    address bufferp;			/* Pointer to item's buffer */
    ulong *rtnlenp;			/* Pointer to return length */
    };

#define MAX_EQVLEN	255		/* Max length of equivalence string */

/*****************************
 * Module Local Own Storage: *
 *****************************/

readonly static $DESCRIPTOR( d_lnm$default, "LNM$FILE_DEV");

/**********************************************
 * Global Procedure and Function definitions: *
 **********************************************/
status SYS$TRNLNM();			/* Translate logical name service */

status LIB$GET_VM();			/* Get dynamic virtual memory */
status LIB$FREE_VM();			/* Release dynamic virtual memory */
status LIB$ANALYZE_SDESC();		/* Analyze string descriptor */
status LIB$SCOPY_R_DX();		/* String copy routine */


/*+ LIB_TRNLNM
 Provide a simple (no item list) interface to $TRNLNM to translate a logical
 name, usually one with a simple equivalence string (a single item).  

   status.wlc.v = LIB_TRNLNM( lognam.rt.dx, eqvnam.wt.dx [,index.rl.r]
                              [,sattrs.rlu.r] [,tabnam.rt.dx]
                              [,outlen.ww.r] [,attrs.wlu.r] )

  status    completion status.  Returned by value.

  lognam    logical name to be translated.  Passed by descriptor.

  eqvnam    returns the equivalence string for the logical name.
            Passed by descriptor.

  index     optional index number (0..127) of the desired equivalence
            name.  Passed by reference.  The default is 0.

  sattrs    optional mask of atrributes controlling the search for the
            logical name (such as LNM$M_CASE_BLIND).  Passed by reference.

  tabnam    optional table name in which to search for the logical name.
            Passed by descriptor.  Default is "LNM$FILE_DEV".

  outlen    optional word in which the length of the equivalence string
            is returned.  Passed by reference.  A zero is returned if
            there is no equivalence name at the specified index.

  attrs     optional longword in which a mask of attributes associated
            with the logical name and equivalence string is returned.
            Passed by reference.

 More information can be found in the documentation on the $TRNLNM system
 service.
-*/

#define	NARG_INDEX	3		/* # of optional argument */
#define	NARG_SATTRS	4
#define	NARG_TABNAM	5
#define NARG_OUTLEN	6
#define NARG_ATTRS	7

status LIB_TRNLNM( lognamp, eqvnamp, indexp, sattrp, tabnamp, outlenp,
		   attrsp )
dscptr lognamp;				/* Descriptor for logical name */
dscptr eqvnamp;				/* Descriptor for equivalence name */
unsigned long *indexp;			/* Pointer to index number */
unsigned long *sattrp;			/* Pointer to search attributes */
dscptr tabnamp;				/* Descriptor for table name */
unsigned short *outlenp;		/* Pointer to return string length */
unsigned long *attrsp;			/* Pointer to return attributes */
{
int nargs = *((long *) &lognamp - 1);	/* Number of arguments in call */
register status sts;
register unsigned long *sap = NULL;	/* Search attributes pointer */
struct {
    struct an_item index;		/* Equivalence name index */
    struct an_item attributes;		/* Name/equivalence attributes */
    struct an_item estring;		/* Equivalence string item */
    unsigned long list_end;		/* End of item list */
    } item_list;
unsigned short elength;			/* Length of equivalence string */
$DESCRIPTOR0( d_lognam);		/* Prototype logical name descriptor */
$DESCRIPTOR0( d_table );		/* Prototype table name descriptor */

/*
 Initialize the item list.
*/
item_list.index.buflen = item_list.attributes.buflen = 4;
item_list.estring.buflen = MAX_EQVLEN;
item_list.attributes.bufferp = item_list.index.bufferp = 0;
item_list.index.rtnlenp = item_list.attributes.rtnlenp = NULL;
item_list.estring.rtnlenp = &elength;
item_list.index.itmcod = LNM$_INDEX;
item_list.attributes.itmcod = LNM$_ATTRIBUTES;
item_list.estring.itmcod = LNM$_STRING;
item_list.list_end = 0;

/*
 Check for optional table name or use default of LNM$FILE_DEV
*/
if ((nargs >= NARG_TABNAM) && (tabnamp != NULL))
    {
    sts = LIB$ANALYZE_SDESC( tabnamp, &d_table.dsc$w_length,
					&d_table.dsc$a_pointer);
    RETURN_ON_FAILURE( sts);
    }
else
    d_table = d_lnm$default;

/*
 Check for optional logical name search attributes.
*/
if ((nargs >= NARG_SATTRS) && (sattrp != NULL))
    sap = sattrp;			/* Use caller's search attributes */

/*
 Check for optional return of name/translation attributes.
*/
if ((nargs >= NARG_ATTRS) && (attrsp != NULL))
    item_list.attributes.bufferp = (address) attrsp;
else
    item_list.attributes.bufferp = (address) &item_list.attributes.bufferp;

/*
 Check for optional equivalence name index.
*/
if ((nargs >= NARG_INDEX) && (indexp != NULL))
    item_list.index.bufferp = (address) indexp;
else
    item_list.index.bufferp = (address) &item_list.index.rtnlenp;

/*
 Fixup descriptor (fixed-length) for logical name.
*/
sts = LIB$ANALYZE_SDESC( lognamp, &d_lognam.dsc$w_length,
				&d_lognam.dsc$a_pointer);
RETURN_ON_FAILURE( sts);

/*
 Get temporary buffer for equivalence name.
*/
sts = LIB$GET_VM( &MAX_EQVLEN, &item_list.estring.bufferp);
RETURN_ON_FAILURE( sts);

/*
 Get name translation using the item list.
*/
sts = SYS$TRNLNM( sap, &d_table, &d_lognam, 0, &item_list);
if ($VMS_SUCCESS( sts))
    {
    sts = LIB$SCOPY_R_DX( &elength, item_list.estring.bufferp, eqvnamp);
    if ($VMS_FAILURE( sts))
	{
	(void) LIB$FREE_VM( &MAX_EQVLEN, &item_list.estring.bufferp);
	return sts;
	}

    /*
     Check for optional return of length of equivalence name.
    */
    if ((nargs >= NARG_OUTLEN) && (outlenp != NULL))
	*outlenp = elength;

    sts = LIB$FREE_VM( &MAX_EQVLEN, &item_list.estring.bufferp);
    }
else
    (void) LIB$FREE_VM( &MAX_EQVLEN, &item_list.estring.bufferp);

return sts;
}

/*+ _TRNLNM_MAX_IND
 Provide a simple (no item list) interface to $TRNLNM to return the
 maximum equivalence string index for a logical name.

   status.wlc.v = LIB_TRNLNM_MAX_INDEX( lognam.rt.dx, max_index.wl.r
                              [,sattrs.rlu.r] [,tabnam.rt.dx] )

  status    completion status.  Returned by value.

  lognam    logical name to be used.  Passed by descriptor.

  max_index longword in which the largest equivalence name index is
            returned.  Passed by reference.  If no equivalence names
            exist, then a -1 is returned.

  sattrs    optional mask of atrributes controlling the search for the
            logical name (such as LNM$M_CASE_BLIND).  Passed by reference.

  tabnam    optional table name in which to search for the logical name.
            Passed by descriptor.  Default is "LNM$FILE_DEV".

 More information can be found in the documentation on the $TRNLNM system
 service.
-*/

#define	NARG2_SATTRS	3		/* # of optional argument */
#define	NARG2_TABNAM	4

status LIB_TRNLNM_MAX_INDEX( lognamp, max_indexp, sattrp, tabnamp )
dscptr lognamp;				/* Descriptor for logical name */
unsigned long *max_indexp;		/* Pointer to return max index */
unsigned long *sattrp;			/* Pointer to search attributes */
dscptr tabnamp;				/* Descriptor for table name */
{
int nargs = *((long *) &lognamp - 1);	/* Number of arguments in call */
register unsigned long *sap = NULL;	/* Search attributes pointer */
register status sts;
struct {
    struct an_item max_index;		/* Maximum index value item */
    unsigned long list_end;		/* End of item list */
    } item_list;
$DESCRIPTOR0( d_lognam);		/* Prototype logical name descriptor */
$DESCRIPTOR0( d_table );		/* Prototype table name descriptor */

/*
 Initialize the item list.
*/
item_list.max_index.buflen = 4;
item_list.max_index.bufferp = max_indexp;
item_list.max_index.rtnlenp = 0;
item_list.max_index.itmcod = LNM$_MAX_INDEX;
item_list.list_end = 0;

/*
 Check for optional table name or use default of LNM$FILE_DEV
*/
if ((nargs >= NARG2_TABNAM) && (tabnamp != NULL))
    {
    sts = LIB$ANALYZE_SDESC( tabnamp, &d_table.dsc$w_length,
					&d_table.dsc$a_pointer);
    RETURN_ON_FAILURE( sts);
    }
else
    d_table = d_lnm$default;

/*
 Check for optional logical name search attributes.
*/
if ((nargs >= NARG2_SATTRS) && (sattrp != NULL))
    sap = sattrp;			/* Use caller's search attributes */

/*
 Fixup descriptor (fixed-length) for logical name.
*/
sts = LIB$ANALYZE_SDESC( lognamp, &d_lognam.dsc$w_length,
				&d_lognam.dsc$a_pointer);
RETURN_ON_FAILURE( sts);

return SYS$TRNLNM( sap, &d_table, &d_lognam, 0, &item_list);
}
