#module  FORT_ENTAB  "V2.0"

/*
 Facility: 
    General VAX utility program

 Abstract:
    Replace spaces with tab characters in FORTRAN source files according
    to the tabbing rules associated with VAX and PDP-11 FORTRAN.

 Environment:
    Foreign DCL command.

 Author: Frank J. Nagy		Fermilab Research Division EED/Controls
 Modification History:

 V1.0	01-Aug-84  FJN	Created
 V2.0	07-Sep-86  FJN	Modified for VAX native mode using VAX C V2.x
*/

#ifdef vaxc
#include stdio
#else
#include <stdio.h>
#endif

#define	FALSE	0
#define	TRUE	1

#define	BLANK	' '
#define	TAB	'\t'
#define	RETURN	'\r'
#define	NEWLINE	'\n'
#define	EOS	'\0'
#define QUOTE	'\''

static char	line[133];		/* Character line buffer */
static char	tabstops[133];		/* Tab stop indicators */

static int	tflag = FALSE;		/* Set TRUE if "-x" option given to use
					   tabs for single spaces where valid */
static int	xflag = FALSE;		/* Set TRUE if "-x" option given to
					   eliminate columns >72 of input */

extern char	*strchr();		/* String search function */

#define	CONTCOL		5		/* Continuation column in "line" */
#define	FIRSTCOL	6		/* 1st code column in "line" */
#define TOSSCOL		72		/* First column tossed by "-x" */


main(argc, argv)
char *argv[];
int argc;
{
register int	i, j;			/* Counters */
register char	*lend;			/* General string pointer */
long	lnum;				/* Statement label holder */
int	label();			/* Label test routine */
long	linecount;			/* Count of lines processes */

lend = NULL;
/*
  Process the command line arguments by re-opening the standard
  input and output files.  Also handle the option flags.
*/
for (i = 1; i < argc; i++)
    if (argv[i][0] == '-')
	for (j = 1; argv[i][j] != EOS; j++)
	    if (tolower(argv[i][j]) == 't')
		tflag++;		/* Set flag for "-t" */
	    else if (tolower(argv[i][j]) == 'x')
		xflag++;		/* Set flag for "-x" */
	    else
		fprintf( stderr, "Unknown option '%c'\n", argv[i][j]);

    else if (lend == 0)
	{
	if (freopen(argv[i], "r", stdin) == NULL)
	    {
	    perror(argv[i]);
	    exit(1);
	    }
	lend++;
	}
    else if (lend == 1)
	{
#ifdef vaxc
	if (freopen(argv[i], "w", stdout, "rfm=var", "rat=cr") == NULL)
#else
	if (freopen(argv[i], "w", stdout) == NULL)
#endif
	    {
	    perror(argv[i]);
	    exit(1);
	    }
	lend++;
	}

/*
  Fill in the tab stop indicator tables.
*/
for (i = 0; i < sizeof tabstops; i++)
    tabstops[i] = i/8;

i = 0;					/* Continuation counter */
linecount = 0;				/* Line counter */
/*
  Read input file line-by-line until EOF.
*/
while (gets(line) != NULL)
    {
    linecount++;
    /*
      If "-x" used, then make sure input line string stops after
      column 72.
    */
    if (xflag)
	line[TOSSCOL] = EOS;

    /*
      Get pointer to end of string and work backwards over
      trailing spaces and tabs.
    */
    for (lend = line; *lend != EOS; lend++);
    if (lend != line)
	{
	lend--;
	while ((*lend == BLANK) || (*lend == TAB))
	    lend--;
	*(++lend) = EOS;
	}

    /*
      Now handle the rules for tab insertion specific to
      PDP-11 and VAX FORTRAN sources.
    */
    switch (line[0])
	{
	case EOS:			/* Blank line */
	printf( "\n");
	break;

	case 'C':			/* Comment line */
	case 'c':
	case 'D':
	case 'd':
	case '!':
	NormalRules();
	break;

	default:			/* FORTRAN statement */
	if (line[CONTCOL] != BLANK)
	    {
	    if ((line[CONTCOL] < '1') || (line[CONTCOL] > '9'))
		printf( "\t%c", '1' + (char) (i++ & 07));
	    else
		printf( "\t%c", line[5]);
					/* A continuation line statement */
	    StatementRules( &line[FIRSTCOL]);
	    }
	else if (label( &lnum))
	    {
	    printf( "%ld\t", lnum);	/* Label then TAB, and rest...  */
	    StatementRules( &line[FIRSTCOL]);
	    i = 0;			/* Reset continuation index */
	    }
	else
	    {
	    printf( "\t");		/* Preceed statement by a TAB */
	    StatementRules( &line[FIRSTCOL]);
	    i = 0;			/* Reset continuation index */
	    }
	break;
	}
    }

fprintf( stderr, "  %ld lines processed\n", linecount);
}		

/*
  The following function implements normal tab insertion rules for
  a line of text (tabs spaced every 8 columns).
*/
NormalRules()
{
register int	col;		/* Scan column number */
register int	new;		/* Output Column number */
register int	last;		/* Last non-blank column number */

for (col = new = last = 0; line[col] != EOS; col++)
    if (line[col] != BLANK)
	if (col == last)
	    line[new++] = line[last++];
	else
	    {
	    while (last < col)
		if (tabstops[col] == tabstops[last])
		    line[new++] = line[last++];
		else if ((col == (last + 1)) && !tflag)
		    line[new++] = line[last++];
		else
		    {
		    line[new++] = '\t';
		    last = 8*(tabstops[last] + 1);
		    }
	    line[new++] = line[last++];
	    }

line[new] = EOS;
/*
  Output the tabbed line.
*/
printf( "%s\n", line);
}

/*
  The following function implements tab insertion rules for
  a FORTRAN statements starting with column 7 (tabs every 8
  columns starting with column 17).  Watch for quotes to disable
  the tabbing rules within paired quotes.
*/
StatementRules( statement )
char	statement[];		/* Remainder of FORTRAN statement */
{
register int	col;		/* Scan column number */
register int	new;		/* Output Column number */
register int	last;		/* Last non-blank column number */
register int	qflag;		/* Flags when inside quoted text */

qflag = FALSE;

for (new = col = last = 0; statement[col] != EOS; col++)
    if (qflag)
	{
	if ((statement[new++] = statement[last++]) == QUOTE)
	    qflag = FALSE;			/* End quote seen */
	}
    else if (statement[col] != BLANK)
	if (col == last)
	    qflag = (statement[new++] = statement[last++]) == QUOTE;
	else
	    {
	    while (last < col)			/* Process multiple spaces */
		if (tabstops[col] == tabstops[last])
		    statement[new++] = statement[last++];
		else if ((col == (last + 1)) && !tflag)
		    statement[new++] = statement[last++]; /* Leave 1 space */
		else 
		    {
		    statement[new++] = '\t';		/* Tab goes in... */
		    last = 8*(tabstops[last] + 1);
		    }
	    qflag = (statement[new++] = statement[last++]) == QUOTE;
	    }

statement[new] = EOS;

/*
  Output the tabbed statement.
*/
printf( "%s\n", statement);
}

/*
  Routine to test for and decode a FORTRAN statement label on the line.
*/
label( num)
long	*num;				/* To return the label number */
{
char ltext[6];				/* To hold the label text */
long atol();				/* Convert text to binary number */
int strcmp();				/* String comparison */

strncpy( ltext, line, CONTCOL);
ltext[CONTCOL] = EOS;
/*
  If we have some non-blank characters, assume they constitue a normal
  FORTRAN label (a number) and convert it to binary.
*/
if (strcmp( ltext, "     ") != 0)
    {
    *num = atol( ltext);
    return (TRUE);
    }
else
    return (FALSE);
}
