/*--- EDIT # 0003	28-MAY-1986 13:54   SYS$SYSROOT:[RVT.UTILS]FGETSS.C;4 */
/*--- Previous edit 	27 Jan 1983   11:55:14	DB1:[21,22]FGETSS.C;4  */
/*
.title	fgetss	Read a string from a file
;+
;
; Index		Read a string from a file
;
; Usage
;
;	char *fgetss(buffer, maxbytes, fp);
;
;	char *fgets(buffer, maxbytes, fp);
;
; In:
;	int	maxbytes
;	FILE	*fp;
;
; Out:
;	char	*buffer;
;
; Description
;
;	Fgets() reads a string into the buffer from the indicated file.
;	Maxbytes is the maximum number of bytes to read.
;	The string is terminated by a newline character,
;	which is followed by a null.  Fgets() normally returns buffer.
;	It returns NULL on end of file or error.
;
;	Fgetss() is identical to fgets() except that the terminating
;	newline is removed.
;
; Status:
;	
; Bugs
;
;	The buffer should be at least 2 characters longer than
;	"maxbytes", to allow for newline and NULL.
;
;	Note that fgets keeps the newline, while fgetss() and gets()
;	delete it. To delete the newline after executing fgets(),
;	execute the following:
;
;		buffer[strlen(buffer) - 1] = 0;
;
;-
*/
/* Mod 5-86 by RVT   Handle err/eof properly
*/

#include <stdio.h>
char *fgetss(buffer, maxbytes, fp)
register char	*buffer;
register int	maxbytes;
FILE	*fp;
{
   register char c;
   char *orig;		/* original buffer start */

   orig = buffer;
   while ((--maxbytes > 0) && (c = getc(fp),(!feof(fp)))) {
      if (c == '\n')		/* quit on newline */
	 break;
      *buffer++ = c;
   }
   *buffer = NULL;		/* trailing null */
   return ((char *)(feof(fp) ? NULL : orig));
}
