/*--- EDIT # 0009	9 Apr 1984   18:59:14	DB2:[310,61]XRF0.C;14  */
/*--- PREVIOUS EDIT	9 Apr 1984   18:54:58	DB2:[310,61]XRF0.C;13  */
#
/*
 *
 *
 * The  information  in  this  document  is  subject  to  change
 * without  notice  and  should not be construed as a commitment
 * by Digital Equipment Corporation or by DECUS.
 * 
 * Neither Digital Equipment Corporation, DECUS, nor the authors
 * assume any responsibility for the use or reliability of  this
 * document or the described software.
 * 
 * 	Copyright (C) 1980, DECUS
 * 
 * 
 * General permission to copy or modify, but not for profit,  is
 * hereby  granted,  provided that the above copyright notice is
 * included and reference made to  the  fact  that  reproduction
 * privileges were granted by DECUS.
 *
*/

/*
 *              ***************
 *              * X R F 0 . C *
 *              ***************
 *
 * This is the mainline program for the C cross referencer and lister.
 * It handles the following tasks:
 *
 *      - Scans the command line
 *      - Opens the appropriate files
 *      - Calls some initialization functions
 *      - Invokes line by line processing
 *      - Prints the cross-reference index
 *      - Finishes up processing as appropriate
 *
 *
 * Usage:
 *
 *	xrf [-s] [-l] [-x] [-n] [-o out_file] in_files
 *
 *	Flags:
 *		-s	Spool out_file to printer (RSX11M only (??))
 *		-l	No listing.
 *		-n	Narrow (80 column) output, normal is 132 column.
 *		-x	No XREF
 *		-o file	Write output to the indicated file.  If -o is not
 *			specified, output will be to the first in_file
 *			with the filetype set to ".x"
 *
 * Written By:
 *              Bob Denny
 *              Creative System Design Co.
 *              3452 E. Foothill Blvd. << Suite 601 >>
 *              Pasadena, Ca.  91107
 *              (213) 355-6836
 *
 * Experimental Version         X1.0             5-May-80
 * Split off initialization     X1.1             7-May-80
 * Prototype.                   Y1.2             9-May-80
 * Release.                     V1.3             9-May-80
 * Chg. for DECUS OTS. & Cmplr. V1.4             1-Jul-80
 * Add support for RT-11        V1.5             3-Jul-80
 * Conditionalized spool call	V1.6 MM		 9-Jul-80
 * Added narrow, outfile, etc.	V1.7 MM		10-Jul-80
 * Added multiple infiles	V1.8 MM		22-Jul-80
 * Fspool now in the library	V1.9 MM		 2-Aug-80
 * Added -x & -l switches	RVT		 9 Apr 1984
 *
 *****************************************************************************
*/

#include <stdio.h>                              /* RSX-11M Std. I/O Def's */
#include "xrf.h"                                /* Global definitions */

int pageno = 0;                                 /* Current listing page no. */
int dflag = 0;
int lflag = 0;		/* No listing flag (0 = print listing) */
int xflag = 0;		/* No XREF list flag (0 = print listing) */


main ( argc, argv )
char *argv[];
{                                            /* Start the main prog. */
   register int i;                              /* Arg count */
   register int c;                              /* Switch character code */
   register char *p;                            /* Fast arg pointer */
   char *in_file;                               /* Source file name string */
   char *out_file;				/* Cref file name string */
   int splflg;                                  /* Spool file flag */
   int nofiles;					/* Flag "got a file" */

    rperline = (80 - 16)/RSIZE;	/* Set references (default is "-n") */
   out_file = NULL;				/* No output file yet */
   splflg = 0;                                  /* Default no spool */
   for ( i=1; i<argc; ++i )                      /* For each token */
      {
      p = argv[i];                              /* p --> i'th arg token */
      if ( *p++ == '-' )                         /* if it's a switch token */
	 {
	 while ( c = *p++ )                      /* Each char is a switch */
	    switch (tolower(c))                  /* Process 'em */
	 {
	 case 'd':			/* Announce each file name */
	    dflag++;
	    break;

	 case 'x':
	    xflag++;
	    break;

	 case 'l':
	    lflag++;
	    break;

	 case 'o':			/* Output file */
	    argv[i] = 0;
	    if (++i >= argc)
	       useage();
	    out_file = argv[i];
	    argv[i] = 0;
	    goto nextarg;

	 case 'n':			/* Narrow output */
	    rperline = (80 - 16)/RSIZE;	/* Set references */
	    break;

	 case 's':                        /* Spool listing */
	    splflg++;
	    break;

	    default :                        /* Bad switch, give help */
		 useage();
	 }
	 argv[i] = 0;				/* Drop this one */
      }
      nextarg:;
   }                                         /* Done with cmd string */


   nofiles = 1;
   for ( i = 1; i < argc; i++)
      {
      if ((in_file = argv[i]) == 0)
	 continue;
      initinfile(in_file);
      while (nextinfile()) {
 	 if (nofiles) {
	    initoutfile(out_file);
	    nofiles = 0;
	 }
	 pageno = 0;                            /* Init. page number */
	 lineno = 0;				/* Init. line number */
	 myfree();				/* Clear out storage */
	 root = NULL;                           /* No tree yet */
	 newpage();                             /* Start first page */
	 while (fgets(scanbf, LWIDTH, src) != NULL) /* Get a line of source */
	    {
	    ++lineno;                           /* Increment line number */
	    scanp = scanbf;                     /* Reset scan pointer */
	    if (!xflag)
	       while (getid())                      /* scanp --> id or NULL */
	       root = search(idbuf, root);      /* Process identifier */
	    /* Till done with this line */
	    if (!lflag)
	       lstline();                          /* List this line */
	 }                                   /* Till end of source file */
	    if (!lflag)
	 newpage();                             /* Start index on new page */
	 prtree(root);                          /* Print the index */
      }
   }
   if (nofiles)
      useage();
#ifdef rsx
   if (splflg)                                   /* If we're to spool, */
      fspool(lst);                              /* Cross your fingers. */
#endif
   exit();                                      /* Exit closing files */
}                                            /***** END MAIN *****/

/*
 *
 * Listing control routines. Newpage also used by prtree.
 *
*/

newpage()                                       /* Start new page */
{
   ++pageno;                                    /* Increment page number */
   linpg = 0;                                   /* Reset line-in-page count */
   fprintf(lst, "%s%d\n\n", pghead, pageno);     /* Output page heading */
}

lstline()                                       /* Write listing line out */
{
   if (++linpg > MAXLIN)
      newpage();
   fprintf(lst, "%4d:\t%s", lineno, scanbf);
}

